import os
from contextlib import contextmanager
from pathlib import Path
from typing import Union


@contextmanager
def chdir(path: Union[str, Path]) -> None:
    """Change the current working directory to the given path and revert back on exit."""
    old_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)



import os
import pathlib
from zipfile import ZipFile

import pytest
from click.testing import CliRunner

from fme_packager.cli import pack
from fme_packager.exception import (
    TransformerPythonCompatError,
    CustomTransformerPythonCompatError,
)
from fme_packager.metadata import TransformerMetadata, FormatMetadata, FMEPackageMetadata
from fme_packager.operations import parse_formatinfo
from fme_packager.packager import (
    FMEPackager,
    is_valid_python_compatibility,
    get_formatinfo,
    get_format_visibility,
)


CWD = pathlib.Path(__file__).parent.resolve()


@pytest.mark.parametrize(
    "version, expected_is_valid",
    [
        ("26", False),
        ("27", False),
        ("ArcGISDesktop", False),
        ("34", False),
        ("35", True),
        ("36", True),
        ("37", True),
        ("3.8.9", False),  # 3.x.x is invalid version syntax for PYTHON_COMPATIBILITY
        ("2or3", False),  # only valid for custom transformers
    ],
)
def test_is_valid_python_compatibility(version, expected_is_valid):
    assert is_valid_python_compatibility(version) == expected_is_valid


def test_get_formatinfo(tmp_path):
    package_metadata = FMEPackageMetadata(
        {
            "uid": "package",
            "publisher_uid": "example",
            "package_content": {"formats": [{"name": "myformat"}]},
        }
    )
    format_metadata = FormatMetadata({"name": "myformat"})

    with pytest.raises(ValueError):
        filepath = tmp_path / "emptyformat.db"
        with open(filepath, "w") as f:
            f.write(";")
        get_formatinfo(package_metadata, format_metadata, filepath)

    with pytest.raises(ValueError):
        filepath = tmp_path / "formatnotfound.db"
        with open(filepath, "w") as f:
            f.write(
                "SAFE.CKAN.CKANDATASTORE|CKAN DataStore|NONE|BOTH|NONE|NO||NON_SPATIAL|NO|YES|YES|YES|3|3|CKAN|NO|CKAN"
            )
        get_formatinfo(package_metadata, format_metadata, filepath)

    filepath = tmp_path / "myformat.db"
    with open(filepath, "w") as f:
        f.writelines(
            "EXAMPLE.PACKAGE.MYFORMAT|My Format|NONE|BOTH|NONE|NO||NON_SPATIAL|NO|YES|YES|YES|3|3|MYFORMAT|NO|MYFORMAT|Coordinates\n"
            + "EXAMPLE.PACKAGE.NOTMYFORMAT|Not My Format|NONE|BOTH|NONE|NO||NON_SPATIAL|NO|YES|YES|YES|3|3|MYFORMAT|NO|MYFORMAT|Coordinates,3D"
        )

    formatinfo = get_formatinfo(package_metadata, format_metadata, filepath)
    assert formatinfo.FORMAT_NAME == "EXAMPLE.PACKAGE.MYFORMAT"


def test_get_format_visibility():
    base_formatinfo = parse_formatinfo(
        "EXAMPLE.PACKAGE.MYFORMAT|My Format|NONE|BOTH|NONE|NO||NON_SPATIAL|NO|YES|YES|YES|3|3|MYFORMAT|NO|MYFORMAT|Coordinates"
    )

    assert get_format_visibility(base_formatinfo._replace(DIRECTION="BOTH", VISIBLE="NO")) == ""
    assert get_format_visibility(base_formatinfo._replace(DIRECTION="BOTH", VISIBLE="YES")) == "rw"

    assert get_format_visibility(base_formatinfo._replace(DIRECTION="BOTH", VISIBLE="INPUT")) == "r"
    for visibility in ["YES", "INPUT"]:
        assert (
            get_format_visibility(base_formatinfo._replace(DIRECTION="INPUT", VISIBLE=visibility))
            == "r"
        )

    assert (
        get_format_visibility(base_formatinfo._replace(DIRECTION="BOTH", VISIBLE="OUTPUT")) == "w"
    )
    for visibility in ["YES", "OUTPUT"]:
        assert (
            get_format_visibility(base_formatinfo._replace(DIRECTION="OUTPUT", VISIBLE=visibility))
            == "w"
        )

    with pytest.raises(ValueError):
        assert get_format_visibility(base_formatinfo._replace(DIRECTION="INPUT", VISIBLE="OUTPUT"))


@pytest.mark.parametrize(
    "transformer_path,metadata,expected_exc",
    [
        ("valid_package/transformers/MyGreeter.fmx", None, None),
        (
            "valid_package/transformers/MyGreeter.fmx",
            {"name": "TheirGreeter", "version": 1},
            "Name must be",
        ),
        ("incompatible_package/transformers/MyGreeter.fmx", None, TransformerPythonCompatError),
        (
            "incompatible_custom_package/transformers/epochToTimestamp29.fmx",
            None,
            CustomTransformerPythonCompatError,
        ),
        ("custom_package/transformers/customFooBar.fmx", None, None),
        ("custom_package/transformers/epochToTimestamp29.fmx", None, None),
        (
            "custom_package/transformers/customEncrypted2Ver.fmx",
            {"name": "test", "version": 2},
            'Custom transformer Insert Mode must be "Linked Always"',
        ),
        ("fmxj_package/transformers/DemoGreeter.fmxj", None, None),
    ],
)
def test_validate_transformer(transformer_path, metadata, expected_exc):
    """
    Load and validate a transformer.
    If expected_exc is given and is an exception, it's expected to be raised.
    If it's a string, it's expected to be a substring in the raised exception message.
    The package.yml in the transformer's parent folder is automatically loaded.
    The metadata argument overrides the transformer's entry in package.yml.
    If the transformer has no entry in package.yml, the metadata must be provided.
    """
    transformer_abs_path = CWD / "fixtures" / transformer_path
    packager = FMEPackager(transformer_abs_path.parent.parent)
    if metadata:
        metadata = TransformerMetadata(metadata)
    else:
        try:
            metadata = {i.name: i for i in packager.metadata.transformers}[
                transformer_abs_path.stem
            ]
        except KeyError as e:
            raise KeyError(
                f"{transformer_abs_path.stem} not in package.yml, but no metadata given"
            ) from e

    if not expected_exc:
        # Expect success
        packager.validate_transformer(transformer_abs_path, metadata)
    elif isinstance(expected_exc, str):
        # Expect a keywords in exception message
        with pytest.raises(Exception) as exc:
            packager.validate_transformer(transformer_abs_path, metadata)
        assert expected_exc in str(exc.value)
    else:
        # Expect a particular exception class
        with pytest.raises(expected_exc):
            packager.validate_transformer(transformer_abs_path, metadata)


@pytest.mark.parametrize("package_name", ["valid_package", "fmxj_package"])
def test_pack(package_name):
    runner = CliRunner()
    result = runner.invoke(pack, [str(CWD / "fixtures" / package_name)])
    assert result.exit_code == 0

    # Sanity check: expect FPKG to exist and contain package.yml at root level
    dist_dir = CWD / "fixtures" / package_name / "dist"
    fpkg_name = os.listdir(dist_dir)[0]
    with ZipFile(dist_dir / fpkg_name) as z:
        assert "package.yml" in z.namelist()



import os

from fme_packager.utils import chdir


def test_chdir(tmp_path):
    old_dir = os.getcwd()

    new_dir = tmp_path / "new_dir"
    new_dir.mkdir()

    with chdir(new_dir):
        assert os.getcwd() == str(new_dir)

    assert os.getcwd() == old_dir



import shutil
import tempfile
from json import dumps as json_dumps

from fme_packager.operations import valid_fpkg_file, zip_filename_for_fpkg
from fme_packager.packager import FMEPackager


class FMEVerifier:
    def __init__(self, file, verbose=False, output_json=False):
        self.file = file
        self.verbose = verbose
        self.output_json = output_json

    def verify(self):
        success, message = True, "valid"
        try:
            self._unzip_and_build()
        except Exception as e:
            success, message = False, str(e)

        if self.output_json:
            result = json_dumps({"status": "success" if success else "error", "message": message})
        else:
            result = "Success: Package Valid" if success else f"Error Validating Package: {message}"

        return result

    def _unzip_and_build(self):
        self.file = valid_fpkg_file(self.file)

        # Create a temporary directory
        with tempfile.TemporaryDirectory() as temp_dir:
            # Change the extension of the file to .zip
            temp_zip_file = str(zip_filename_for_fpkg(temp_dir, self.file))
            self._print(f"Creating temporary zip file {temp_zip_file}")
            shutil.copy(self.file, temp_zip_file)

            # Unpack the zip file
            shutil.unpack_archive(temp_zip_file, temp_dir)

            # Verify the fpkg files by building the package
            steps = FMEPackager(temp_dir, self.verbose)
            steps.build()
            steps.make_fpkg()

    def _print(self, msg):
        if self.verbose:
            print(msg)



"""
Tools for extracting key information out of FME's transformer definition files.
"""

import json
import os.path
import re
from abc import ABC, abstractmethod
from collections import namedtuple


class Transformer(ABC):
    """Represents one version of a transformer."""

    @property
    @abstractmethod
    def name(self):
        pass

    @property
    @abstractmethod
    def version(self):
        pass

    @property
    @abstractmethod
    def python_compatibility(self):
        pass

    @property
    @abstractmethod
    def categories(self):
        pass

    @property
    @abstractmethod
    def aliases(self):
        pass

    @property
    @abstractmethod
    def visible(self):
        pass


NamedTransformerHeader = namedtuple(
    "NamedTransformerHeader",
    "name version category guid insert_mode blocked_looping process_count process_group_by process_groups_ordered build_num preserves_attrs deprecated pyver",
)


def parse_custom_transformer_header(line):
    """
    Parses custom transformer header.

    :param str line: Custom transformer header line from FMX.
    :return: Parsed header
    """
    fields = line.replace("# TRANSFORMER_BEGIN", "").strip().split(",")
    header = NamedTransformerHeader(*fields[: len(NamedTransformerHeader._fields)])
    return header._replace(version=int(header.version), build_num=int(header.build_num))


class CustomTransformer(Transformer):
    def __init__(self, lines):
        super().__init__()
        self.header: NamedTransformerHeader = None
        self.lines = lines
        for line in lines:
            if line.startswith(b"# TRANSFORMER_BEGIN"):
                self.header = parse_custom_transformer_header(line.decode("utf8"))
                break
        else:
            raise ValueError("TRANSFORMER_BEGIN line not found")

    @property
    def name(self):
        return self.header.name

    @property
    def version(self):
        return self.header.version

    @property
    def python_compatibility(self):
        return self.header.pyver

    @property
    def categories(self):
        return self._split_prop("category", ",")

    @property
    def aliases(self):
        return []

    @property
    def visible(self):
        return True

    @property
    def is_encrypted(self):
        return self.lines[0].strip() == b"FMW0001"

    def _split_prop(self, property_name, sep=","):
        return (
            [p.strip() for p in getattr(self.header, property_name).split(sep)]
            if getattr(self.header, property_name)
            else []
        )


class FmxTransformer(Transformer):
    def __init__(self, lines):
        super().__init__()
        self.lines = lines
        self.props = {}
        for line in self.lines:
            match = re.match(r"^(.+?):\s+(.+?)$", line)
            if match:
                name = match.group(1).strip()
                if name.startswith("PARAMETER_"):
                    continue
                self.props[name] = match.group(2).strip()
        if not self.name or not self.version:
            raise ValueError("TRANSFORMER_NAME or VERSION not found")

    @property
    def name(self):
        return self.props.get("TRANSFORMER_NAME")

    @property
    def version(self):
        return int(self.props.get("VERSION"))

    @property
    def python_compatibility(self):
        return self.props.get("PYTHON_COMPATIBILITY")

    def _split_prop(self, property_name, sep=","):
        return (
            [p.strip() for p in self.props.get(property_name).split(sep)]
            if self.props.get(property_name)
            else []
        )

    @property
    def categories(self):
        return self._split_prop("CATEGORY", ",")

    @property
    def aliases(self):
        return self._split_prop("ALIASES", " ")

    @property
    def visible(self):
        return self.props.get("VISIBLE", "yes").lower() == "yes"


class FmxjTransformer(Transformer):
    def __init__(self, info, version_def):
        self.info = info
        self.json_def = version_def

    @property
    def name(self):
        return self.info["name"]

    @property
    def version(self):
        return self.json_def["version"]

    @property
    def python_compatibility(self):
        # FIXME: key typo from tstportConfiguration/testdata/PortConfiguration.fmxj
        return self.json_def.get("pythonCompatability")

    @property
    def categories(self):
        return self.info.get("categories", [])

    @property
    def aliases(self):
        return self.info.get("aliases", [])

    @property
    def visible(self):
        return not self.info.get("deprecated", False)


class TransformerFile(ABC):
    """Represents a transformer file containing one or more transformer versions."""

    def __init__(self, file_path):
        self.file_path = file_path

    @abstractmethod
    def versions(self):
        pass


def get_matching_indexes(lines, matcher):
    """
    Returns indexes of lines where line is (or can be decoded as) UTF-8 and matcher function is true.
    """
    indexes = []
    for i, line in enumerate(lines):
        if not isinstance(line, str):
            try:
                line = line.decode("utf8")
            except (UnicodeDecodeError, AttributeError):
                continue
        if matcher(line):
            indexes.append(i)
    return indexes


class FmxFile(TransformerFile):
    def __init__(self, file_path):
        super().__init__(file_path)

    def versions(self):
        with open(self.file_path, "r") as f:
            lines = f.readlines()
        transformer_begin_indexes = get_matching_indexes(
            lines, lambda l: l.startswith("TRANSFORMER_NAME")
        )
        for def_num, i in enumerate(transformer_begin_indexes):
            if i == transformer_begin_indexes[-1]:
                end_idx = len(lines) - 1
            else:
                end_idx = transformer_begin_indexes[def_num + 1] - 1
            yield FmxTransformer(lines[i:end_idx])


class CustomTransformerFmxFile(TransformerFile):
    def __init__(self, file_path):
        super().__init__(file_path)

    def versions(self):
        with open(self.file_path, "rb") as f:
            lines = f.readlines()
        transformer_begin_indexes = get_matching_indexes(
            lines, lambda l: l.startswith("# TRANSFORMER_BEGIN")
        )
        for def_num, i in enumerate(transformer_begin_indexes):
            if i == transformer_begin_indexes[-1]:
                end_idx = len(lines) - 1
            else:
                end_idx = transformer_begin_indexes[def_num + 1] - 2
            yield CustomTransformer(lines[i - 1 : end_idx])


class FmxjFile(TransformerFile):
    def __init__(self, file_path):
        super().__init__(file_path)
        with open(file_path) as f:
            self.body = json.load(f)

    def versions(self):
        for item in self.body["versions"]:
            yield FmxjTransformer(self.body["info"], item)


def load_transformer(transformer_path) -> TransformerFile:
    filename, ext = os.path.splitext(transformer_path)
    ext = ext.lower()
    if ext == ".fmx":
        # Check first line for Custom Transformer text
        with open(transformer_path, "rb") as f:
            line = f.readline()
        if line.startswith(b"#!") or line.startswith(b"FMW0001"):
            return CustomTransformerFmxFile(transformer_path)
        return FmxFile(transformer_path)
    if ext == ".fmxj":
        return FmxjFile(transformer_path)
    raise ValueError(f"Unrecognized transformer: {transformer_path}")



import os
import pathlib

from fme_packager.transformer import (
    load_transformer,
    CustomTransformerFmxFile,
    CustomTransformer,
    FmxFile,
    FmxjFile,
)


CWD = pathlib.Path(__file__).parent.resolve()


def test_custom_transformer(custom_package_dir):
    fmx = load_transformer(os.path.join(custom_package_dir, "transformers", "customFooBar.fmx"))
    assert isinstance(fmx, CustomTransformerFmxFile)
    defs = list(fmx.versions())
    assert len(defs) == 1
    for item in defs:
        assert isinstance(item, CustomTransformer)
        assert item.header.insert_mode == '"Linked Always"'
        assert item.name == "example.my-package.customFooBar"
        assert item.version == 1
        assert item.python_compatibility == "37"
        assert item.categories == ["Attributes"]
        assert item.aliases == []


def test_custom_transformer_encrypted_versioned(custom_package_dir):
    fmx = load_transformer(
        os.path.join(custom_package_dir, "transformers", "customEncrypted2Ver.fmx")
    )
    assert isinstance(fmx, CustomTransformerFmxFile)
    defs = list(fmx.versions())
    assert len(defs) == 2
    for i, item in enumerate(defs):
        assert isinstance(item, CustomTransformer)
        assert item.header.insert_mode == '"Embedded Always"'
        assert item.name == "example.my-package.test"
        assert item.version == len(defs) - i
        assert item.python_compatibility == "310"
        assert item.is_encrypted == (i == 0)  # Latest version is encrypted


def test_fmx_transformer(valid_package_dir):
    fmx = load_transformer(os.path.join(valid_package_dir, "transformers", "MyGreeter.fmx"))
    assert isinstance(fmx, FmxFile)
    defs = list(fmx.versions())
    assert len(defs) == 1
    item = defs[0]
    assert item.name == "example.my-package.MyGreeter"
    assert item.version == 1
    assert item.python_compatibility == "36"
    assert item.categories == ["Integrations"]
    assert item.aliases == [
        "example.my-package.HerGreeter",
        "example.my-package.TheirGreeter",
        "example.my-package.OurGreeter",
    ]


def test_fmxj_transformer(valid_package_dir):
    fmx = load_transformer(CWD / "fixtures" / "fmxj_package" / "transformers" / "DemoGreeter.fmxj")
    assert isinstance(fmx, FmxjFile)
    defs = list(fmx.versions())
    assert len(defs) == 1
    item = defs[0]
    assert item.name == "example.my-package.DemoGreeter"
    assert item.version == 1
    assert not item.python_compatibility
    assert item.categories == ["Integrations"]
    assert item.aliases == ["example.my-package.NemoGreeter", "example.my-package.MemoGreeter"]



"""
fme_packager command line interface.
"""

import inspect
import os
import shutil
import sys
import sysconfig
from pathlib import Path

import click
from cookiecutter.main import cookiecutter

import fme_packager
from fme_packager.packager import FMEPackager
from fme_packager.summarizer import summarize_fpkg
from fme_packager.verifier import FMEVerifier


@click.group()
@click.version_option(package_name="fme_packager")
def cli():
    """
    fme_packager: The FME Packages Tool.
    """
    pass


COOKIECUTTER_TEMPLATES = {
    "transformer": "https://github.com/safesoftware/fpkg-transformer-template/archive/refs/heads/main.zip",
}


@cli.command()
@click.argument("template", type=click.Choice(sorted(COOKIECUTTER_TEMPLATES.keys())))
def init(template):
    """
    Initialize a FME Package from a template.

    The template is initialized in a subdirectory of the current directory.

    TEMPLATE -- Name of the template to use.
    """
    print("Enter the values to use for the template. Press Enter to accept the [default].")
    cookiecutter(COOKIECUTTER_TEMPLATES[template])


@cli.command()
@click.argument("help_path", type=click.Path(exists=True, file_okay=True))
@click.argument("fpkg_path", type=click.Path(exists=True, file_okay=False, writable=True))
def apply_help(help_path, fpkg_path):
    """
    Import a Safe TechPubs doc export into an FME Package directory.

    This operation also converts package_aliases.flali to package_help.csv at the destination.

    HELP_PATH -- Path to a ZIP or directory of a Safe TechPubs doc export. Read only.

    FPKG_PATH -- Path to the FME Package root. Its 'help' subdirectory will be recreated.
    """
    steps = FMEPackager(fpkg_path)
    steps.apply_help(help_path)


@cli.command()
@click.argument("path", type=click.Path(exists=True, file_okay=False, writable=True))
def pack(path):
    """
    Create an .fpkg file.

    Package contents are validated during this process.
    Components not referenced by the package's metadata.yml may not be included in the resulting fpkg.

    PATH -- Path to an FME Package directory.
    """
    steps = FMEPackager(path)
    steps.build()
    steps.make_fpkg()


@cli.command()
@click.argument("file", type=click.Path(exists=False))
@click.option("--verbose", "-v", is_flag=True, help="Show build steps")
@click.option("--json", is_flag=True, help="Output result as JSON")
def verify(file, verbose, json):
    """
    Verify that a .fpkg file is valid.

    Package contents are validated during this process.

    FILE -- Path to an FME .fpkg package file.
    """
    verifier = FMEVerifier(file, verbose, json)
    result = verifier.verify()
    print(result)


@cli.command()
@click.argument("file", type=click.Path(exists=True, file_okay=True, dir_okay=False))
def summarize(file):
    """
    Extract a JSON representation of an .fpkg file.

    FILE -- Path to an FME .fpkg package file.
    """
    result = summarize_fpkg(file)
    print(result)


@cli.command()
@click.option(
    "--fme-home",
    prompt=True,
    default=lambda: os.environ.get("FME_HOME"),
    show_default="FME_HOME environment variable",
    type=click.Path(exists=True, file_okay=False, writable=True),
    required=True,
    help="Path to FME installation",
)
@click.option(
    "--site-packages-dir",
    default=lambda: sysconfig.get_paths()["purelib"],
    show_default="current Python environment",
    type=click.Path(exists=True, file_okay=False, writable=True),
    required=True,
    help="Path to Python environment's site-packages",
)
def config_env(fme_home, site_packages_dir):
    """
    Configure a Python environment for access to fmeobjects
    and the Python libraries included with FME.
    """
    leaf_dir = "python%s%s" % (sys.version_info.major, sys.version_info.minor)
    paths_to_add = [
        "#" + fme_home,
        os.path.join(fme_home, "python"),
        os.path.join(
            fme_home,
            "python",
            f"fme-plugins-py{sys.version_info.major}{sys.version_info.minor}.zip",
        ),
        os.path.join(fme_home, "python", leaf_dir),
        os.path.join(fme_home, "fmeobjects", leaf_dir),
    ]

    dst_pth = os.path.join(site_packages_dir, "fme_env.pth")
    print("Writing " + dst_pth)
    with open(dst_pth, "w") as f:
        for path in paths_to_add:
            f.write(path + "\n")
        f.write("import fme_env\n")

    dst_py = os.path.join(site_packages_dir, "fme_env.py")
    print("Writing " + dst_py)
    src = Path(inspect.getfile(fme_packager)).parent / "fme_env.py"
    shutil.copy(src, dst_py)

    print("\nThis Python environment is now set up for access to FME and fmeobjects.")
    print("If the FME install location changes, re-run this script to update paths.")


if __name__ == "__main__":
    cli()



import json
import os
import shutil
import tempfile
from collections import namedtuple
from pathlib import Path
from typing import Iterable, List

import yaml
from jsonschema import validate
from jsonschema.exceptions import ValidationError

from fme_packager.operations import valid_fpkg_file, zip_filename_for_fpkg, parse_formatinfo
from fme_packager.packager import _load_format_line
from fme_packager.transformer import load_transformer, TransformerFile, Transformer
from fme_packager.utils import chdir


def _unpack_fpkg_file(directory: str, fpkg_file: str):
    """
    Unpack the FME Package file.

    :param directory: The directory where the FME Package file will be unpacked.
    :param fpkg_file: The FME Package file to be unpacked.
    :return: The directory where the FME Package file was unpacked.
    """
    valid_fpkg = valid_fpkg_file(fpkg_file)
    zip_file = zip_filename_for_fpkg(directory, valid_fpkg)
    shutil.copy(valid_fpkg, zip_file)
    shutil.unpack_archive(zip_file, directory)

    return directory


def _manifest_path(fpkg_dir: str) -> Path:
    """
    Return the manifest file.

    :param fpkg_dir: The directory where the FME Package file was unpacked.
    :return: The path to the manifest file.
    """
    return Path(fpkg_dir) / "package.yml"


def _parsed_manifest(yaml_file: Path) -> dict:
    """
    Parse the manifest file.

    :param yaml_file: The path to the manifest yaml file.
    :return: A dictionary of the parsed YAML.
    """
    with open(yaml_file, "r") as file:
        return yaml.safe_load(file)


TransformerFilenames = namedtuple(
    "TransformerFilenames", ["filename", "readme_filename"], defaults=[None, None]
)

FormatFilenames = namedtuple(
    "FormatFilenames", ["filename", "readme_filename", "db_filename"], defaults=[None, None, None]
)


def _transformer_filenames(transformer_name: str) -> TransformerFilenames:
    """
    Retrieve filenames for the transformer and its readme

    Input dict must have the key:
    - 'name': The name of the transformer.

    :param transformer_name: The name of the transformer to retrieve filenames for.
    :return: A tuple containing the filename and readme filename.
    """
    if not transformer_name:
        return TransformerFilenames(filename=None, readme_filename=None)

    result = dict()
    for ext, key in [("fmx", "filename"), ("fmxj", "filename"), ("md", "readme_filename")]:
        potential_filename = str(Path("transformers") / f"{transformer_name}.{ext}")
        if os.path.exists(potential_filename):
            result[key] = potential_filename

    return TransformerFilenames(**result)


def _format_filenames(format_name: str) -> FormatFilenames:
    """
    Retrieve filenames for the format and its readme

    :param format_name: The name of the format to retrieve filenames for.
    :return: A tuple containing the filename, db filename and readme filename.
    """
    if not format_name:
        return FormatFilenames(filename=None, readme_filename=None, db_filename=None)

    filenames = {
        "filename": str(Path("formats") / f"{format_name}.fmf"),
        "db_filename": str(Path("formats") / f"{format_name}.db"),
        "readme_filename": str(Path("formats") / f"{format_name}.md"),
    }

    for key, filename in filenames.items():
        if not os.path.exists(filename):
            filenames[key] = None

    return FormatFilenames(**filenames)


def _transformer_data(loaded_file: TransformerFile) -> dict:
    """
    Get the relevant data from the loaded transformer file.

    Output dict will have the following keys:
    - 'versions': A list of dictionaries, each containing:
        - 'name': The name of the transformer.
        - 'version': The version number of the transformer.
        - 'categories': The categories of the transformer.
        - 'aliases': The aliases of the transformer.
        - 'visible': The visibility status of the transformer.

    :param loaded_file: The loaded transformer file.
    :return: The updates for the transformer dictionary.
    """
    versions: list[Transformer] = loaded_file.versions()
    return {
        "versions": [
            {
                "name": version.name,
                "version": version.version,
                "categories": version.categories,
                "aliases": version.aliases,
                "visible": version.visible,
            }
            for version in versions
        ],
    }


def _content_description(readme_filename: str) -> dict:
    """
    Get the description of the transformer or format from the readme file.

    Output dict will have keys:
    - 'description': The description of the transformer or format.
    - 'description_format': The format of the description either 'md', 'text', or 'html'.

    :param readme_filename: The filename of the readme file.
    :return: The update for the transformer or format dictionary.
    """
    try:
        with open(readme_filename, "r") as file:
            return {
                "description": file.read(),
                "description_format": "md",
            }
    except OSError:
        return {
            "description": None,
            "description_format": None,
        }


def _get_all_categories(transformers: Iterable[dict], formats: Iterable[dict]) -> List[str]:
    """
    Get the union of all the categories from the transformers.

    :param transformers: An iterable of transformer dicts
    :return: An alphabetically sorted list of all categories from the transformers.
    """
    all_categories = set()
    for transformer in transformers:
        if not transformer.get("versions", None):
            continue
        highest_version = max(
            transformer["versions"], key=lambda transformer_version: transformer_version["version"]
        )
        if not highest_version["categories"]:
            continue
        all_categories.update(highest_version["categories"])

    for format in formats:
        if not format.get("categories", None):
            continue
        all_categories.update(format["categories"])

    all_categories_list = list(all_categories)
    all_categories_list.sort()
    return all_categories_list


def _enhance_transformer_info(transformers: Iterable[dict]) -> Iterable[dict]:
    """
    Enhance the transformer entries in the manifest with additional information.

    :param transformers: An iterable of transformer dicts
    :return: An iterable of transformer dicts with enhanced information
    """
    if not transformers:
        return []

    for transformer in transformers:
        filenames = _transformer_filenames(transformer["name"])
        loaded_transformer = load_transformer(filenames.filename)
        transformer.update(_transformer_data(loaded_transformer))
        transformer.update(_content_description(filenames.readme_filename))
        transformer["latest_version"] = transformer.pop("version")

    return transformers


def _to_bool(value: str) -> bool:
    """
    Convert a string to a boolean.

    :param value: The value to convert.
    :return: The boolean value.
    """
    return value.upper() == "YES"


def _format_data(format_line: str) -> dict:
    """
    Get the relevant data from the loaded format file.

    Output dict will have the following keys:
    - 'visible': The visibility status of the format.
    - 'fds_info': The FDS info line of the format.
    - 'categories': The categories of the format.

    :param format_line: The loaded format line.
    :return: The updates for the format dictionary.
    """

    format_info = parse_formatinfo(format_line)
    format_data = {"visible": None, "fds_info": None, "categories": None}

    if format_info:
        format_data["fds_info"] = format_line
        format_data["visible"] = _to_bool(format_info.VISIBLE)
        format_data["categories"] = (
            format_info.FORMAT_CATEGORIES.split(",") if format_info.FORMAT_CATEGORIES else []
        )

    return format_data


def _enhance_format_info(publisher_uid: str, uid: str, formats: Iterable[dict]) -> Iterable[dict]:
    """
    Enhance the format entries in the manifest with additional information.

    :param publisher_uid: The publisher UID
    :param uid: The UID
    :param formats: An iterable of format dicts
    :return: An iterable of format dicts with enhanced information
    """
    if not formats:
        return []

    for format in formats:
        filenames = _format_filenames(format["name"])
        format_data = _format_data(_load_format_line(filenames.db_filename))
        format["short_name"] = format["name"]
        format["name"] = f"{publisher_uid}.{uid}.{format['name']}"
        format.update(_content_description(filenames.readme_filename))
        format.update(format_data)

    return formats


def _load_output_schema() -> dict:
    """
    Load the output schema for the summarizer in jsonschema format.

    :return: The output schema.
    """
    with open(Path(__file__).parent / "summarizer_spec.json", "r") as file:
        return json.load(file)


def summarize_fpkg(fpkg_path: str) -> str:
    """
    Summarize the FME Package.

    The output conforms to summarizer_spec.json.

    :param fpkg_path: The path to the FME Package.
    :return: A JSON string of the summarized FME Package, or an error message under the key 'error'.
    """
    output_schema = _load_output_schema()

    with tempfile.TemporaryDirectory() as temp_dir:
        _unpack_fpkg_file(temp_dir, fpkg_path)

        with chdir(temp_dir):
            manifest = _parsed_manifest(_manifest_path(temp_dir))
            transformers = manifest.get("package_content", {}).get("transformers", [])
            formats = manifest.get("package_content", {}).get("formats", [])
            manifest["package_content"] = manifest.get("package_content", {})
            manifest["package_content"]["transformers"] = _enhance_transformer_info(transformers)
            manifest["package_content"]["formats"] = _enhance_format_info(
                manifest.get("publisher_uid", ""), manifest.get("uid", ""), formats
            )
            manifest["categories"] = _get_all_categories(transformers, formats)
        try:
            validate(manifest, output_schema)
        except ValidationError as e:
            return json.dumps(
                {
                    "status": "error",
                    "message": f"The generated output did not conform to the schema: {e.message}",
                },
                indent=2,
            )

        return json.dumps(manifest, indent=2)



import json
import pathlib
import tempfile
from unittest.mock import patch

import pytest
import yaml
from click.testing import CliRunner
from pathlib import Path

from fme_packager import summarizer
from fme_packager.cli import summarize
from fme_packager.summarizer import TransformerFilenames, FormatFilenames
from tests.conftest import mock_transformer, mock_transformer_file

CWD = pathlib.Path(__file__).parent.resolve()


@pytest.mark.parametrize(
    "transformer_name, expected",
    [
        # Case 1: Transformer files do not exist
        (
            "non_existing_transformer",
            TransformerFilenames(filename=None, readme_filename=None),
        ),
        # Case 2: 'name' key is null
        (
            None,
            TransformerFilenames(filename=None, readme_filename=None),
        ),
        # Case 3: Transformer .fmx file exists but .fmxj file does not
        (
            "fmx_transformer",
            TransformerFilenames(
                filename="transformers/fmx_transformer.fmx",
                readme_filename="transformers/fmx_transformer.md",
            ),
        ),
        # Case 4: Transformer .fmxj file exists but .fmx file does not
        (
            "fmxj_transformer",
            TransformerFilenames(
                filename="transformers/fmxj_transformer.fmxj",
                readme_filename="transformers/fmxj_transformer.md",
            ),
        ),
    ],
)
def test__transformer_filenames(transformer_name, expected, mocker):
    def side_effect(path):
        if "non_existing_transformer" in path:
            return False
        if "fmx_transformer" in path and path.endswith(".fmx"):
            return True
        if "fmxj_transformer" in path and path.endswith(".fmxj"):
            return True
        if path.endswith(".md"):
            return True
        return False

    mocker.patch("os.path.exists", side_effect=side_effect)

    result = summarizer._transformer_filenames(transformer_name)
    assert result == expected


@pytest.mark.parametrize(
    "format_name, exists, expected",
    [
        (
            "test_format",
            True,
            FormatFilenames(
                filename=str(Path("formats") / "test_format.fmf"),
                db_filename=str(Path("formats") / "test_format.db"),
                readme_filename=str(Path("formats") / "test_format.md"),
            ),
        ),
        (
            "non_existing_format",
            False,
            FormatFilenames(
                filename=None,
                db_filename=None,
                readme_filename=None,
            ),
        ),
    ],
)
def test__format_filenames(format_name, exists, expected):
    with patch("os.path.exists", return_value=exists):
        result = summarizer._format_filenames(format_name)
        assert result == expected


def test__transformer_data(mock_transformer_file, mock_transformer):
    result = summarizer._transformer_data(mock_transformer_file)

    assert result["versions"] == [
        {
            "name": mock_transformer.name,
            "version": mock_transformer.version,
            "categories": mock_transformer.categories,
            "aliases": mock_transformer.aliases,
            "visible": mock_transformer.visible,
        }
    ]


def test__format_data():
    mock_format_line = (
        "SAFE.AIRTABLE.AIRTABLE|Airtable|NONE|BOTH|BOTH|NO||NON_SPATIAL|NO|YES|YES|YES|1|1|Airtable|NO"
        "|Airtable|3D,Coordinates"
    )

    result = summarizer._format_data(mock_format_line)
    assert result == {
        "visible": True,
        "fds_info": mock_format_line,
        "categories": ["3D", "Coordinates"],
    }


@pytest.fixture
def mock_transformers(mocker):
    # Mock the TransformerFile object
    mock_transformer_file1 = mocker.Mock()
    mock_transformer_file1.versions.return_value = [mocker.Mock(categories={"cat1", "cat2"})]

    mock_transformer_file2 = mocker.Mock()
    mock_transformer_file2.versions.return_value = [mocker.Mock(categories=None)]

    mock_transformer_file3 = mocker.Mock()
    mock_transformer_file3.versions.return_value = [mocker.Mock(categories={"cat3", "cat4"})]

    mock_transformer_file4 = mocker.Mock()
    mock_transformer_file4.versions.return_value = [mocker.Mock(categories={"cat4", "cat5"})]

    return [
        {
            "versions": [
                {"version": 1, "categories": {"cat6", "cat7"}},
                {"version": 2, "categories": {"cat1", "cat2"}},
            ]
        },
        {"versions": []},
        {"versions": [{"version": 1, "categories": {"cat3", "cat4"}}]},
        {"versions": [{"version": 1, "categories": {"cat4", "cat5"}}]},
    ]


@pytest.fixture
def mock_formats():
    return [
        {"categories": {"cat-f1", "cat3"}},
        {"categories": {"cat-f2", "cat4"}},
    ]


def test__get_all_categories(mock_transformers, mock_formats):
    result = summarizer._get_all_categories(mock_transformers, mock_formats)
    assert result == ["cat-f1", "cat-f2", "cat1", "cat2", "cat3", "cat4", "cat5"]


def test__add_content_description(mocker):
    transformer = {"readme_filename": "MyGreeter.md"}
    mocker.patch("builtins.open", mocker.mock_open(read_data="Test Description"))
    result = summarizer._content_description(transformer["readme_filename"])
    assert result["description"] == "Test Description"
    assert result["description_format"] == "md"


def test__parsed_manifest():
    # Create a temporary file
    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".yml") as temp:
        # Write some data to the file
        data = {
            "key1": "value1",
            "key2": "value2",
            "key3": "value3",
        }
        yaml.dump(data, temp)

    # Call the function with the path of the temporary file
    result = summarizer._parsed_manifest(pathlib.Path(temp.name))

    # Assert that the returned dictionary matches the data we wrote to the file
    assert result == data


@pytest.mark.parametrize(
    "fpkg_path, expected_output_path",
    [
        (
            CWD / "fixtures" / "fpkgs" / "example.my-package-0.1.0.fpkg",
            CWD / "fixtures" / "json_output" / "summarize_example.my-package-0.1.0.fpkg.json",
        ),
        (
            CWD / "fixtures" / "fpkgs" / "example.my-format-0.1.0.fpkg",
            CWD / "fixtures" / "json_output" / "summarize_example.my-format-0.1.0.fpkg.json",
        ),
    ],
)
def test_summarize_fpkg(fpkg_path, expected_output_path):
    expected_output = json.load(open(expected_output_path))
    result = json.loads(summarizer.summarize_fpkg(str(fpkg_path)))
    assert result == expected_output


def test_summarize_empty_fpkg(monkeypatch):
    monkeypatch.setattr(
        summarizer,
        "_parsed_manifest",
        lambda x: {
            "fpkg_version": 1,
            "uid": "my-package",
            "publisher_uid": "example",
            "name": "My FME Package",
            "description": "A short description of my FME Package.",
            "version": "0.1.0",
            "minimum_fme_build": 19238,
            "author": {"name": "G Raymond", "email": "me@example.com"},
        },
    )
    fpkg_path = CWD / "fixtures" / "fpkgs" / "example.my-package-0.1.0.fpkg"
    expected_output = json.load(
        open(CWD / "fixtures" / "json_output" / "summarize_example.empty.json")
    )
    result = json.loads(summarizer.summarize_fpkg(str(fpkg_path)))
    assert result == expected_output


def test_summarize():
    runner = CliRunner()
    fpkg_path = str(CWD / "fixtures" / "fpkgs" / "example.my-package-0.1.0.fpkg")
    result = runner.invoke(summarize, [fpkg_path])
    assert result.exit_code == 0
    json.loads(result.output)
